@@ -1,6 +1,6 @@
 /*
 
-  $Id: gsm-sms.c,v 1.159 2007/01/06 22:53:21 pkot Exp $
+  $Id: gsm-sms.c,v 1.184 2008/09/10 21:00:18 pkot Exp $
 
   G N O K I I
 
@@ -32,14 +32,21 @@
 
 */
 
+#include "config.h"
+
 #include <stdlib.h>
 #include <string.h>
+#include <time.h>
 
 #include "gnokii-internal.h"
 #include "gnokii.h"
 
 #include "sms-nokia.h"
 
+#ifdef ENABLE_NLS
+#  include <locale.h>
+#endif
+
 #undef ERROR
 #define ERROR() do { if (error != GN_ERR_NONE) return error; } while (0)
 
@@ -50,21 +57,6 @@ struct sms_udh_data {
 
 #define MAX_SMS_PART 140
 
-/* These are positions in the following (headers[]) table. */
-#define SMS_UDH_CONCATENATED_MESSAGES  1
-#define SMS_UDH_RINGTONES              2
-#define SMS_UDH_OPERATOR_LOGOS         3
-#define SMS_UDH_CALLER_LOGOS           4
-#define SMS_UDH_MULTIPART_MESSAGES     5
-#define SMS_UDH_WAP_VCARD              6
-#define SMS_UDH_WAP_CALENDAR           7
-#define SMS_UDH_WAP_VCARDSECURE        8
-#define SMS_UDH_WAP_VCALENDARSECURE    9
-#define SMS_UDH_VOICE_MESSAGES        10
-#define SMS_UDH_FAX_MESSAGES          11
-#define SMS_UDH_EMAIL_MESSAGES        12
-#define SMS_UDH_WAP_PUSH              13
-
 /* User data headers */
 static struct sms_udh_data headers[] = {
 	{ 0x00, "" },
@@ -153,20 +145,28 @@ static char *sms_timestamp_print(u8 *number)
 	/* Ugly hack, but according to the GSM specs, the year is stored
          * as the 2 digit number. */
 	if ((10*(number[0] & 0x0f) + (number[0] >> 4)) < 70)
-		sprintf(buffer, "20");
+		snprintf(buffer, sizeof(buffer), "20");
 	else
-		sprintf(buffer, "19");
+		snprintf(buffer, sizeof(buffer), "19");
 
 	for (i = 0; i < 6; i++) {
 		int c;
 		char buf2[4];
 		switch (i) {
-		case 0: case 1: c = '-'; break;
-		case 3: case 4: c = ':'; break;
-		default: c = ' '; break;
+		case 0:
+		case 1:
+			c = '-';
+			break;
+		case 3:
+		case 4:
+			c = ':';
+			break;
+		default:
+			c = ' ';
+			break;
 		}
 		snprintf(buf2, 4, "%d%d%c", number[i] & 0x0f, number[i] >> 4, c);
-		strcat(buffer, buf2);
+		strncat(buffer, buf2, sizeof(buffer) - strlen(buffer));
 	}
 
 	/* The GSM spec is not clear what is the sign of the timezone when the
@@ -174,11 +174,13 @@ static char *sms_timestamp_print(u8 *number)
 	 * some are not. If your operator does use incompatible SMSC and wrong
 	 * sign disturbs you, change the sign here.
 	 */
-	if (number[6] & 0x08) strcat(buffer, "-");
-	else strcat(buffer, "+");
+	if (number[6] & 0x08)
+		strncat(buffer, "-", sizeof(buffer) - strlen(buffer));
+	else
+		strncat(buffer, "+", sizeof(buffer) - strlen(buffer));
 	/* The timezone is given in quarters. The base is GMT. */
-	sprintf(buf, "%02d00", (10 * (number[6] & 0x07) + (number[6] >> 4)) / 4);
-	strcat(buffer, buf);
+	snprintf(buf, sizeof(buf), "%02d00", (10 * (number[6] & 0x07) + (number[6] >> 4)) / 4);
+	strncat(buffer, buf, sizeof(buffer) - strlen(buffer));
 
 	return buffer;
 #undef LOCAL_DATETIME_MAX_LENGTH
@@ -197,16 +199,20 @@ static char *sms_timestamp_print(u8 *number)
  */
 gn_timestamp *sms_timestamp_unpack(unsigned char *number, gn_timestamp *dt)
 {
-	if (!dt) return NULL;
+	if (!dt)
+		return NULL;
 	memset(dt, 0, sizeof(gn_timestamp));
-	if (!number) return dt;
+	if (!number)
+		return dt;
 
-	dt->year     =  10 * (number[0] & 0x0f) + (number[0] >> 4);
+	dt->year = 10 * (number[0] & 0x0f) + (number[0] >> 4);
 
 	/* Ugly hack, but according to the GSM specs, the year is stored
          * as the 2 digit number. */
-	if (dt->year < 70) dt->year += 2000;
-	else dt->year += 1900;
+	if (dt->year < 70)
+		dt->year += 2000;
+	else
+		dt->year += 1900;
 
 	dt->month    =  10 * (number[1] & 0x0f) + (number[1] >> 4);
 	dt->day      =  10 * (number[2] & 0x0f) + (number[2] >> 4);
@@ -221,7 +227,8 @@ gn_timestamp *sms_timestamp_unpack(unsigned char *number, gn_timestamp *dt)
 	 * some are not. If your operator does use incompatible SMSC and wrong
 	 * sign disturbs you, change the sign here.
 	 */
-	if (number[6] & 0x08) dt->timezone = -dt->timezone;
+	if (number[6] & 0x08)
+		dt->timezone = -dt->timezone;
 
 	return dt;
 }
@@ -234,14 +241,18 @@ gn_timestamp *sms_timestamp_unpack(unsigned char *number, gn_timestamp *dt)
  */
 unsigned char *sms_timestamp_pack(gn_timestamp *dt, unsigned char *number)
 {
-	if (!number) return NULL;
+	if (!number)
+		return NULL;
 	memset(number, 0, GN_SMS_DATETIME_MAX_LENGTH);
-	if (!dt) return number;
+	if (!dt)
+		return number;
 
 	/* Ugly hack, but according to the GSM specs, the year is stored
          * as the 2 digit number. */
-	if (dt->year < 2000) dt->year -= 1900;
-	else dt->year -= 2000;
+	if (dt->year < 2000)
+		dt->year -= 1900;
+	else
+		dt->year -= 2000;
 
 	number[0]    = (dt->year   / 10) | ((dt->year   % 10) << 4);
 	number[1]    = (dt->month  / 10) | ((dt->month  % 10) << 4);
@@ -257,7 +268,8 @@ unsigned char *sms_timestamp_pack(gn_timestamp *dt, unsigned char *number)
 	 * some are not. If your operator does use incompatible SMSC and wrong
 	 * sign disturbs you, change the sign here.
 	 */
-	if (dt->timezone < 0) number[6] |= 0x08;
+	if (dt->timezone < 0)
+		number[6] |= 0x08;
 
 	return number;
 }
@@ -283,7 +295,8 @@ static gn_error sms_status(unsigned char status, gn_sms *sms)
 	sms->user_data[0].type = GN_SMS_DATA_Text;
 	sms->user_data[1].type = GN_SMS_DATA_None;
 	if (status < 0x03) {
-		strcpy(sms->user_data[0].u.text, _("Delivered"));
+		sms->user_data[0].dr_status = GN_SMS_DR_Status_Delivered;
+		snprintf(sms->user_data[0].u.text, sizeof(sms->user_data[0].u.text), "%s", _("Delivered"));
 		switch (status) {
 		case 0x00:
 			dprintf("SM received by the SME");
@@ -296,11 +309,11 @@ static gn_error sms_status(unsigned char status, gn_sms *sms)
 			break;
 		}
 	} else if (status & 0x40) {
-		strcpy(sms->user_data[0].u.text, _("Failed"));
+		snprintf(sms->user_data[0].u.text, sizeof(sms->user_data[0].u.text), "%s", _("Failed"));
 		/* more detailed reason only for debug */
 		if (status & 0x20) {
 			dprintf("Temporary error, SC is not making any more transfer attempts\n");
-
+			sms->user_data[0].dr_status = GN_SMS_DR_Status_Failed_Permanent;
 			switch (status) {
 			case 0x60:
 				dprintf("Congestion");
@@ -315,7 +328,7 @@ static gn_error sms_status(unsigned char status, gn_sms *sms)
 				dprintf("Service rejected");
 				break;
 			case 0x64:
-				dprintf("Quality of service not aviable");
+				dprintf("Quality of service not available");
 				break;
 			case 0x65:
 				dprintf("Error in SME");
@@ -326,6 +339,7 @@ static gn_error sms_status(unsigned char status, gn_sms *sms)
 			}
 		} else {
 			dprintf("Permanent error, SC is not making any more transfer attempts\n");
+			sms->user_data[0].dr_status = GN_SMS_DR_Status_Failed_Temporary;
 			switch (status) {
 			case 0x40:
 				dprintf("Remote procedure error");
@@ -340,7 +354,7 @@ static gn_error sms_status(unsigned char status, gn_sms *sms)
 				dprintf("Not obtainable");
 				break;
 			case 0x44:
-				dprintf("Quality of service not aviable");
+				dprintf("Quality of service not available");
 				break;
 			case 0x45:
 				dprintf("No internetworking available");
@@ -363,7 +377,8 @@ static gn_error sms_status(unsigned char status, gn_sms *sms)
 			}
 		}
 	} else if (status & 0x20) {
-		strcpy(sms->user_data[0].u.text, _("Pending"));
+		sms->user_data[0].dr_status = GN_SMS_DR_Status_Pending;
+		snprintf(sms->user_data[0].u.text, sizeof(sms->user_data[0].u.text), "%s", _("Pending"));
 		/* more detailed reason only for debug */
 		dprintf("Temporary error, SC still trying to transfer SM\n");
 		switch (status) {
@@ -380,7 +395,7 @@ static gn_error sms_status(unsigned char status, gn_sms *sms)
 			dprintf("Service rejected");
 			break;
 		case 0x24:
-			dprintf("Quality of service not aviable");
+			dprintf("Quality of service not available");
 			break;
 		case 0x25:
 			dprintf("Error in SME");
@@ -390,7 +405,8 @@ static gn_error sms_status(unsigned char status, gn_sms *sms)
 			break;
 		}
 	} else {
-		strcpy(sms->user_data[0].u.text, _("Unknown"));
+		sms->user_data[0].dr_status = GN_SMS_DR_Status_Invalid;
+		snprintf(sms->user_data[0].u.text, sizeof(sms->user_data[0].u.text), "%s", _("Unknown"));
 
 		/* more detailed reason only for debug */
 		dprintf("Reserved/Specific to SC: %x", status);
@@ -580,17 +596,20 @@ static gn_error sms_header_decode(gn_sms_raw *rawsms, gn_sms *sms, gn_sms_udh *u
 	case GN_SMS_MT_Submit:
 		dprintf("Mobile Originated (stored) message:\n");
 		break;
-	case GN_SMS_MT_SubmitSent:
-		dprintf("Mobile Originated (sent) message:\n");
+	case GN_SMS_MT_StatusReport:
+		dprintf("Status Report message:\n");
 		break;
 	case GN_SMS_MT_Picture:
 		dprintf("Picture Message:\n");
 		break;
+	case GN_SMS_MT_TextTemplate:
+		dprintf("Text Template:\n");
+		break;
 	case GN_SMS_MT_PictureTemplate:
 		dprintf("Picture Template:\n");
 		break;
-	case GN_SMS_MT_TextTemplate:
-		dprintf("Text Template:\n");
+	case GN_SMS_MT_SubmitSent:
+		dprintf("Mobile Originated (sent) message:\n");
 		break;
 	default:
 		dprintf("Not supported message type: %d\n", sms->type);
@@ -622,7 +641,7 @@ static gn_error sms_header_decode(gn_sms_raw *rawsms, gn_sms *sms, gn_sms_udh *u
 	/* User Data Header */
 	if (rawsms->udh_indicator & 0x40) { /* UDH header available */
 		dprintf("UDH found\n");
-		sms_udh_decode(rawsms->user_data, udh);
+		return sms_udh_decode(rawsms->user_data, udh);
 	}
 
 	return GN_ERR_NONE;
@@ -643,6 +662,7 @@ static gn_error sms_pdu_decode(gn_sms_raw *rawsms, gn_sms *sms)
 
 	error = sms_header_decode(rawsms, sms, &sms->udh);
 	ERROR();
+	sms->user_data[0].dr_status = GN_SMS_DR_Status_None;
 	switch (sms->type) {
 	case GN_SMS_MT_DeliveryReport:
 		sms_status(rawsms->report_status, sms);
@@ -662,8 +682,9 @@ static gn_error sms_pdu_decode(gn_sms_raw *rawsms, gn_sms *sms)
 			sms->user_data[0].type = GN_SMS_DATA_Bitmap;
 			gn_bmp_sms_read(GN_BMP_PictureMessage, rawsms->user_data,
 					NULL, &sms->user_data[0].u.bitmap);
+#ifdef DEBUG
 			gn_bmp_print(&sms->user_data[0].u.bitmap, stderr);
-
+#endif
 			size = rawsms->user_data_length - 4 - sms->user_data[0].u.bitmap.size;
 			/* Second part is a text */
 			sms->user_data[1].type = GN_SMS_DATA_NokiaText;
@@ -686,12 +707,14 @@ static gn_error sms_pdu_decode(gn_sms_raw *rawsms, gn_sms *sms)
 			gn_bmp_sms_read(GN_BMP_PictureMessage,
 					rawsms->user_data + rawsms->user_data[0] + 7,
 					NULL, &sms->user_data[0].u.bitmap);
+#ifdef DEBUG
 			gn_bmp_print(&sms->user_data[0].u.bitmap, stderr);
+#endif
 		}
 		break;
 	/* Plain text message */
 	default:
-		sms_data_decode(rawsms->user_data + sms->udh.length,        /* Skip the UDH */
+		return sms_data_decode(rawsms->user_data + sms->udh.length, /* Skip the UDH */
 				(unsigned char *)&sms->user_data[0].u.text, /* With a plain text message we have only 1 part */
 				rawsms->length,                             /* Length of the decoded text */
 				rawsms->user_data_length,                   /* Length of the encoded text (in full octets) without UDH */
@@ -711,13 +734,253 @@ static gn_error sms_pdu_decode(gn_sms_raw *rawsms, gn_sms *sms)
  * the highlevel SMS. In data->raw_sms there's SMS read by the phone
  * driver, data->sms is the place for the parsed SMS.
  */
-GNOKII_API gn_error gn_sms_parse(gn_data *data)
+gn_error gn_sms_parse(gn_data *data)
 {
 	if (!data->raw_sms || !data->sms) return GN_ERR_INTERNALERROR;
 	/* Let's assume at the moment that all messages are PDU coded */
 	return sms_pdu_decode(data->raw_sms, data->sms);
 }
 
+/**
+ * gn_sms_pdu2raw - Copy PDU data into gn_sms_raw structure
+ * @rawsms: gn_sms_raw structure to be filled
+ * @pdu: buffer containing PDU data
+ * @pdu_len: length of buffer containing PDU data
+ * @flags: GN_SMS_PDU_* flags to control copying
+ *
+ * This function fills a gn_sms_raw structure copying raw values from
+ * a buffer containing PDU data.
+ */
+gn_error gn_sms_pdu2raw(gn_sms_raw *rawsms, unsigned char *pdu, int pdu_len, int flags)
+{
+#define COPY_REMOTE_NUMBER(pdu, offset) { \
+	l = (pdu[offset] % 2) ? pdu[offset] + 1 : pdu[offset]; \
+	l = l / 2 + 2; \
+	if (l + offset + extraoffset > pdu_len || l > GN_SMS_NUMBER_MAX_LENGTH) { \
+		dprintf("Invalid remote number length (%d)\n", l); \
+		ret = GN_ERR_INTERNALERROR; \
+		goto out; \
+	} \
+	memcpy(rawsms->remote_number, pdu + offset, l); \
+	offset += l; \
+	}
+
+#define COPY_USER_DATA(pdu, offset) { \
+	rawsms->length = pdu[offset++]; \
+	rawsms->user_data_length = rawsms->length; \
+	if (rawsms->udh_indicator) \
+		rawsms->user_data_length -= pdu[offset] + 1; \
+	if (pdu_len - offset > 1000) { \
+		dprintf("Phone gave as poisonous (too short?) reply, either phone went crazy or communication went out of sync\n"); \
+		ret = GN_ERR_INTERNALERROR; \
+		goto out; \
+	} \
+	memcpy(rawsms->user_data, pdu + offset, pdu_len - offset); \
+	}
+
+	gn_error ret = GN_ERR_NONE;
+	unsigned int l, extraoffset, offset = 0;
+	int parameter_indicator;
+
+	if (!flags & GN_SMS_PDU_NOSMSC) {
+		l = pdu[offset] + 1;
+		if (l > pdu_len || l > GN_SMS_SMSC_NUMBER_MAX_LENGTH) {
+			dprintf("Invalid message center length (%d)\n", l);
+			ret = GN_ERR_INTERNALERROR;
+			goto out;
+		}
+		memcpy(rawsms->message_center, pdu, l);
+		offset += l;
+	}
+
+	rawsms->reject_duplicates   = 0;
+	rawsms->report_status       = 0;
+	rawsms->reference           = 0;
+	rawsms->reply_via_same_smsc = 0;
+	rawsms->report              = 0;
+
+	/* NOTE: offsets are taken from subclause 9.2.3 "Definition of the TPDU parameters"
+	   of 3GPP TS 23.040 version 7.0.1 Release 7 == ETSI TS 123 040 V7.0.1 (2007-03)
+	   *not* from the tables in subclause 9.2.2 "PDU Type repertoire at SM-TL" which ignore unused bits
+	*/
+
+	/* NOTE: this function handles only short messages received by the phone.
+	   Since the same TP-MTI value is used both for SC->MS and for MS->SC,
+	   it can't simply shift to get the enum value: (pdu[offset] & 0x03) << 1
+	*/
+
+	/* TP-MTI TP-Message-Type-Indicator 9.2.3.1 */
+	switch (pdu[offset] & 0x03) {
+	case 0x00:
+		rawsms->type = GN_SMS_MT_Deliver;
+		break;
+	case 0x01:
+		rawsms->type = GN_SMS_MT_Submit;
+		break;
+	case 0x02:
+		rawsms->type = GN_SMS_MT_StatusReport;
+		break;
+	case 0x03:
+		dprintf("Reserved TP-MTI found\n");
+		return GN_ERR_INTERNALERROR;
+		break;
+	}
+	switch (rawsms->type) {
+	case GN_SMS_MT_Deliver:
+		dprintf("SMS-DELIVER found\n");
+		/* TP-MMS  TP-More-Messages-to-Send */
+		rawsms->more_messages       = pdu[offset] & 0x04;
+		// bits 3 and 4 of the first octet unused;
+		/* TP-SRI  TP-Status-Report-Indication */
+		rawsms->report_status       = pdu[offset] & 0x20;
+		/* TP-UDHI TP-User-Data-Header-Indication */
+		rawsms->udh_indicator       = pdu[offset] & 0x40;
+		/* TP-RP   TP-Reply-Path */
+		rawsms->reply_via_same_smsc = pdu[offset] & 0x80;
+		offset++;
+		/* TP-OA   TP-Originating-Address */
+		extraoffset = 10;
+		COPY_REMOTE_NUMBER(pdu, offset);
+		/* TP-PID  TP-Protocol-Identifier */
+		rawsms->pid = pdu[offset++];
+		/* TP-DCS  TP-Data-Coding-Scheme */
+		rawsms->dcs = pdu[offset++];
+		/* TP-SCTS TP-Service-Centre-Time-Stamp */
+		memcpy(rawsms->smsc_time, pdu + offset, 7);
+		offset += 7;
+		/* TP-UDL  TP-User-Data-Length */
+		/* TP-UD   TP-User-Data */
+		COPY_USER_DATA(pdu, offset);
+		break;
+	case GN_SMS_MT_Submit:
+		dprintf("SMS-SUBMIT found\n");
+		/* TP-RD   TP-Reject-Duplicates */
+		rawsms->reject_duplicates   = pdu[offset] & 0x04;
+		/* TP-VPF  TP-Validity-Period-Format */
+		rawsms->validity_indicator  = (pdu[offset] & 0x18) >> 3;
+		/* TP-SRR  TP-Status-Report-Request */
+		rawsms->report              = pdu[offset] & 0x20;
+		/* TP-UDHI TP-User-Data-Header-Indication */
+		rawsms->udh_indicator       = pdu[offset] & 0x40;
+		/* TP-RP   TP-Reply-Path */
+		rawsms->reply_via_same_smsc = pdu[offset] & 0x80;
+		offset++;
+		/* TP-MR   TP-Message-Reference */
+		rawsms->reference           = pdu[offset++];
+		/* TP-DA   TP-Destination-Address */
+		extraoffset = 3;
+		COPY_REMOTE_NUMBER(pdu, offset);
+		/* TP-PID  TP-Protocol-Identifier */
+		rawsms->pid = pdu[offset++];
+		/* TP-DCS  TP-Data-Coding-Scheme */
+		rawsms->dcs = pdu[offset++];
+		/* TP-VP   TP-Validity-Period */
+		switch (rawsms->validity_indicator) {
+		case GN_SMS_VP_None:
+			/* nothing to convert */
+			break;
+		case GN_SMS_VP_RelativeFormat:
+			rawsms->validity[0] = pdu[offset++];
+			break;
+		case GN_SMS_VP_EnhancedFormat:
+			/* FALL THROUGH */
+		case GN_SMS_VP_AbsoluteFormat:
+			memcpy(rawsms->validity, pdu + offset, 7);
+			offset += 7;
+			break;
+		default:
+			dprintf("Unknown validity_indicator 0x%02x\n", rawsms->validity_indicator);
+			return GN_ERR_INTERNALERROR;
+		}
+		/* TP-UDL  TP-User-Data-Length */
+		/* TP-UD   TP-User-Data */
+		COPY_USER_DATA(pdu, offset);
+		break;
+	case GN_SMS_MT_StatusReport:
+		dprintf("SMS-STATUS-REPORT found\n");
+		/* TP-MMS  TP-More-Messages-to-Send */
+		rawsms->more_messages       = pdu[offset] & 0x04;
+		/* TP-SRQ  TP-Status-Report-Qualifier */
+		rawsms->report              = pdu[offset] & 0x10;
+		/* TP-UDHI TP-User-Data-Header-Indication */
+		rawsms->udh_indicator       = pdu[offset] & 0x40;
+		offset++;
+		/* TP-MR   TP-Message-Reference */
+		rawsms->reference           = pdu[offset++];
+		/* TP-RA   TP-Recipient-Address */
+		extraoffset = 0;
+		COPY_REMOTE_NUMBER(pdu, offset);
+		/* TP-SCTS TP-Service-Centre-Time-Stamp */
+		memcpy(rawsms->smsc_time, pdu + offset, 7);
+		offset += 7;
+		/* TP-DT   TP-Discharge-Time */
+		memcpy(rawsms->time, pdu + offset, 7);
+		offset += 7;
+		/* TP-ST   TP-Status */
+		rawsms->report              = pdu[offset++];
+		/* TP-PI   TP-Parameter-Indicator */
+		dprintf("TP-Parameter-Indicator: 0x%02x\n", pdu[offset++]);
+		parameter_indicator         = pdu[offset];
+		/* handle the "extension bit" skipping the following octects, if any (see 9.2.3.27 TP-Parameter-Indicator) */
+		while (pdu[offset++] & 0x80)
+			;
+		if (parameter_indicator & 0x01) {
+			/* TP-PID  TP-Protocol-Identifier */
+			rawsms->pid = pdu[offset++];
+		}
+		if (parameter_indicator & 0x02) {
+			/* TP-DCS  TP-Data-Coding-Scheme */
+			rawsms->dcs = pdu[offset++];
+		}
+		if (parameter_indicator & 0x04) {
+			/* TP-UDL  TP-User-Data-Length */
+			/* TP-UD   TP-User-Data */
+			COPY_USER_DATA(pdu, offset);
+		}
+		break;
+	case GN_SMS_MT_Command:
+#if 0
+		dprintf("SMS-COMMAND found\n");
+		/* TP-Message-Type-Indicator */
+		/** decoded above **/
+		offset = 5;
+		/* TP-User-Data-Header-Indication */
+		rawsms->udh_indicator       = pdu[offset] & 0x40;
+		/* TP-Status-Report-Request */
+		rawsms->report_status       = pdu[offset] & 0x08;
+		offset++;
+		/* TP-Message Reference */
+		rawsms->reference           = pdu[offset++];
+		/* TP-Protocol-Identifier */
+		rawsms->pid                 = pdu[offset++];
+		/* TP-Command-Type */
+		dprintf("TP-Command-Type 0x%02x\n", pdu[offset++]);
+		/* TP-Message-Number */
+		dprintf("TP-Message-Number 0x%02x\n", pdu[offset++]);
+		/* TP-Destination-Address */
+	l = (pdu[offset] % 2) ? pdu[offset] + 1 : pdu[offset];
+	l = l / 2 + 2;
+	memcpy(rawsms->remote_number, pdu + offset, l);
+	offset += l;
+		/* TP-Command-Data-Length */
+		dprintf("TP-Command-Data-Length 0x%02x\n", pdu[offset++]);
+		/* TP-Command-Data */
+		dprintf("TP-Command-Data 0x%02x ...\n", pdu[offset++]);
+		goto out;
+		break;
+#endif
+		/* FALL THROUGH */
+	default:
+		dprintf("Unknown PDU type %d\n", rawsms->type);
+		return GN_ERR_INTERNALERROR;
+	}
+
+out:
+	return ret;
+#undef COPY_USER_DATA
+#undef COPY_REMOTE_NUMBER
+}
+
 /**
  * gn_sms_request - High-level function for the explicit SMS reading from the phone
  * @data: GSM data for the phone driver
@@ -727,7 +990,7 @@ GNOKII_API gn_error gn_sms_parse(gn_data *data)
  * phone driver. Note that raw_sms field in the gn_data structure must
  * be initialized
  */
-GNOKII_API gn_error gn_sms_request(gn_data *data, struct gn_statemachine *state)
+gn_error gn_sms_request(gn_data *data, struct gn_statemachine *state)
 {
 	if (!data->raw_sms) return GN_ERR_INTERNALERROR;
 	return gn_sm_functions(GN_OP_GetSMS, data, state);
@@ -746,7 +1009,12 @@ GNOKII_API gn_error gn_sms_get(gn_data *data, struct gn_statemachine *state)
 	gn_error error;
 	gn_sms_raw rawsms;
 
-	if (!data->sms) return GN_ERR_INTERNALERROR;
+	if (!data->sms)
+		return GN_ERR_INTERNALERROR;
+	if (data->sms->number < 0)
+		return GN_ERR_EMPTYLOCATION;
+	if (data->sms->memory_type > GN_MT_LAST)
+		return GN_ERR_INVALIDMEMORYTYPE;
 	memset(&rawsms, 0, sizeof(gn_sms_raw));
 	rawsms.number = data->sms->number;
 	rawsms.memory_type = data->sms->memory_type;
@@ -1259,9 +1527,8 @@ gn_error sms_prepare(gn_sms *sms, gn_sms_raw *rawsms)
 		if (sms->udh.udh[i].type == GN_SMS_UDH_ConcatenatedMessages)
 			sms_concat_header_encode(rawsms, sms->udh.udh[i].u.concatenated_short_message.current_number,
 						sms->udh.udh[i].u.concatenated_short_message.maximum_number);
-	sms_data_encode(sms, rawsms);
 
-	return GN_ERR_NONE;
+	return sms_data_encode(sms, rawsms);
 }
 
 static void sms_dump_raw(gn_sms_raw *rawsms)
@@ -1309,7 +1576,7 @@ GNOKII_API gn_error gn_sms_send(gn_data *data, struct gn_statemachine *state)
 		data->raw_sms->message_center[0] = data->raw_sms->message_center[0] / 2 + 1;
 
 	error = sms_prepare(data->sms, data->raw_sms);
-	ERROR();
+	if (error != GN_ERR_NONE) goto cleanup;
 
 	sms_dump_raw(data->raw_sms);
 	if (data->raw_sms->user_data_length > MAX_SMS_PART) {
@@ -1320,7 +1587,8 @@ GNOKII_API gn_error gn_sms_send(gn_data *data, struct gn_statemachine *state)
 
 	dprintf("Sending\n");
 	error = gn_sm_functions(GN_OP_SendSMS, data, state);
- cleanup:
+cleanup:
+	data->sms->reference = data->raw_sms->reference;
 	free(data->raw_sms);
 	data->raw_sms = NULL;
 	return error;
@@ -1336,7 +1604,7 @@ GNOKII_API int gn_sms_udh_add(gn_sms *sms, gn_sms_udh_type type)
 
 static gn_error sms_send_long(gn_data *data, struct gn_statemachine *state)
 {
-	int i, count, start, copied, total, isConcat = -1;
+	int i, count, start, copied, total, isConcat = -1, max_sms_len = MAX_SMS_PART;
 	gn_sms sms;
 	gn_sms_user_data ud[GN_SMS_PART_MAX_NUMBER];
 	gn_error error = GN_ERR_NONE;
@@ -1354,13 +1622,13 @@ static gn_error sms_send_long(gn_data *data, struct gn_statemachine *state)
 	if (isConcat == -1)
 		isConcat = gn_sms_udh_add(data->sms, GN_SMS_UDH_ConcatenatedMessages);
 
-	/* Count the total length of the octets to be sent */
+	/* Count the total length of the message text octets to be sent */
 	total = 0;
 	i = 0;
 	while (data->sms->user_data[i].type != GN_SMS_DATA_None) {
 		switch (data->sms->dcs.u.general.alphabet) {
 		case GN_SMS_DCS_DefaultAlphabet:
-			total += ((data->sms->udh.length+1) % 8 + data->sms->user_data[i].length) * 7 / 8;
+			total += (data->sms->user_data[i].length * 7 + 7) / 8;
 			break;
 		case GN_SMS_DCS_UCS2:
 			total += (data->sms->user_data[i].length * 2);
@@ -1372,10 +1640,11 @@ static gn_error sms_send_long(gn_data *data, struct gn_statemachine *state)
 		memcpy(&ud[i], &data->sms->user_data[i], sizeof(gn_sms_user_data));
 		i++;
 	}
-	total += (data->sms->udh.length+1);
 
+	/* We need to attach user data header to each part */
+	max_sms_len -= (data->sms->udh.length + 1);
 	/* Count number of SMS to be sent */
-	count = (total + MAX_SMS_PART - 1)/ MAX_SMS_PART;
+	count = (total + max_sms_len - 1)/ max_sms_len;
 	dprintf("Will need %d sms-es\n", count);
 
 	start = 0;
@@ -1388,10 +1657,10 @@ static gn_error sms_send_long(gn_data *data, struct gn_statemachine *state)
 		switch (data->sms->dcs.u.general.alphabet) {
 		case GN_SMS_DCS_DefaultAlphabet:
 			start += copied;
-			if (ud[0].length - start >= (MAX_SMS_PART - (data->sms->udh.length+1)) * 8 / 7) {
-				copied = (MAX_SMS_PART - (data->sms->udh.length+1)) * 8 / 7;
+			if (ud[0].length - start >= max_sms_len * 8 / 7) {
+				copied = max_sms_len * 8 / 7;
 			} else {
-				copied = (ud[0].length - start) % ((MAX_SMS_PART - (data->sms->udh.length+1)) * 8 / 7);
+				copied = (ud[0].length - start) % (max_sms_len * 8 / 7);
 			}
 			memset(&data->sms->user_data[0], 0, sizeof(gn_sms_user_data));
 			data->sms->user_data[0].type = ud[0].type;
@@ -1400,15 +1669,15 @@ static gn_error sms_send_long(gn_data *data, struct gn_statemachine *state)
 			break;
 		case GN_SMS_DCS_UCS2:
 			start += copied;
-			copied = (ud[0].length - start) % ((MAX_SMS_PART - (data->sms->udh.length+1)));
+			copied = (ud[0].length - start) % max_sms_len;
 			/* This is maximum size of what we can send */
 			if (copied == 0) {
-				copied = (MAX_SMS_PART - (data->sms->udh.length+1));
+				copied = max_sms_len;
 			}
-			if (copied > (MAX_SMS_PART - (data->sms->udh.length+1)) / 2) {
+			if (copied > max_sms_len / 2) {
 				copied = copied / 2;
-				if (copied < (MAX_SMS_PART - (data->sms->udh.length+1)) / 2)
-					copied = (MAX_SMS_PART - (data->sms->udh.length+1)) / 2;
+				if (copied < max_sms_len / 2)
+					copied = max_sms_len / 2;
 			}
 			if (i < count - 1 && copied % 2) {
 				copied--;
@@ -1617,3 +1886,154 @@ GNOKII_API void gn_wap_push_init(gn_wap_push *wp)
 	wp->header.charset 	= WAPPush_CHARSET;
 	wp->header.stl 		= 0x00; /* string table length */
 }
+
+static char *status2str(gn_sms_message_status status)
+{
+	switch (status) {
+	case GN_SMS_Unread:
+		return "Unread";
+	case GN_SMS_Sent:
+		return "Sent";
+	case GN_SMS_Unsent:
+		return "Unsent";
+	default:
+		return "Read";
+	}
+}
+
+#define MAX_TEXT_SUMMARY	20
+#define MAX_SUBJECT_LENGTH	25
+#define MAX_DATE_LENGTH		255
+
+/* From snprintf(3) manual */
+static char *allocate(char *fmt, ...)
+{
+	char *str, *nstr;
+	int len, size = 100;
+	va_list ap;
+
+	str = calloc(100, sizeof(char));
+	if (!str)
+		return NULL;
+
+	while (1) {
+		va_start(ap, fmt);
+		len = vsnprintf(str, size, fmt, ap);
+		va_end(ap);
+		if (len >= size) /* too small buffer */
+			size = len + 1;
+		else if (len > -1) /* buffer OK */
+			return str;
+		else /* let's try with larger buffer */
+			size *= 2;
+		nstr = realloc(str, size);
+		if (!nstr) {
+			free(str);
+			return NULL;
+		}
+		str = nstr;
+	}
+}
+
+/* Here we allocate place for the line to append and we append it.
+ * We free() the appended line. */
+#define APPEND(dst, src, size) \
+do { \
+	char *ndst; \
+	int old = size; \
+	size += strlen(src); \
+	ndst = realloc(dst, size + 1); \
+	if (!ndst) { \
+		free(dst); \
+		goto error; \
+	} \
+	dst = ndst; \
+	dst[old] = 0; \
+	strcat(dst, src); \
+	free(src); \
+} while (0)
+
+#define CONCAT(dst, src, size, pattern, ...) \
+do { \
+	src = allocate(pattern, __VA_ARGS__); \
+	if (!src) \
+		goto error; \
+	APPEND(dst, src, size); \
+} while (0);
+
+/* Returns allocated space for mbox-compatible formatted SMS */
+/* TODO:
+ *   o add encoding information
+ *   o support non-text SMS (MMS, EMS, delivery report, ...)
+ */
+GNOKII_API char *gn_sms2mbox(gn_sms *sms, char *from)
+{
+	struct tm t, *loctime;
+	time_t caltime;
+	int size = 0;
+	char *tmp;
+#ifdef ENABLE_NLS
+	char *loc;
+#endif
+	char *buf = NULL, *aux = NULL;
+
+	t.tm_sec = sms->smsc_time.second;
+	t.tm_min = sms->smsc_time.minute;
+	t.tm_hour = sms->smsc_time.hour;
+	t.tm_mday = sms->smsc_time.day;
+	t.tm_mon = sms->smsc_time.month - 1;
+	t.tm_year = sms->smsc_time.year - 1900;
+#ifdef HAVE_TM_GMTON
+	if (sms->smsc_time.timezone)
+		t.tm_gmtoff = sms->smsc_time.timezone * 3600;
+#endif
+	caltime = mktime(&t);
+	loctime = localtime(&caltime);
+
+#ifdef ENABLE_NLS
+	loc = setlocale(LC_ALL, "C");
+#endif
+	CONCAT(buf, tmp, size, "From %s@%s %s", sms->remote.number, from, asctime(loctime));
+
+	tmp = calloc(MAX_DATE_LENGTH, sizeof(char));
+	if (!tmp)
+		goto error;
+	strftime(tmp, MAX_DATE_LENGTH - 1, "Date: %a, %d %b %Y %H:%M:%S %z (%Z)\n", loctime);
+#ifdef ENABLE_NLS
+	setlocale(LC_ALL, loc);
+#endif
+	APPEND(buf, tmp, size);
+
+	CONCAT(buf, tmp, size, "From: %s@%s\n", sms->remote.number, from);
+	CONCAT(buf, tmp, size, "X-GSM-SMSC: %s\n", sms->smsc.number);
+	CONCAT(buf, tmp, size, "X-GSM-Status: %s\n", status2str(sms->status));
+	CONCAT(buf, tmp, size, "X-GSM-Memory: %s\n", gn_memory_type2str(sms->memory_type));
+
+	aux = calloc(16, sizeof(char)); /* assuming location will never have more than 15 digits */
+	if (!aux)
+		goto error;
+	snprintf(aux, 16, "%d", sms->number);
+	CONCAT(buf, tmp, size, "X-GSM-Location: %s\n", aux);
+	free(aux);
+
+	if (strlen(sms->user_data[0].u.text) < MAX_SUBJECT_LENGTH) {
+		CONCAT(buf, tmp, size, "Subject: %s\n\n", sms->user_data[0].u.text);
+	} else {
+		aux = calloc(MAX_TEXT_SUMMARY + 1, sizeof(char));
+		if (!aux)
+			goto error;
+		snprintf(aux, MAX_TEXT_SUMMARY, "%s", sms->user_data[0].u.text);
+		CONCAT(buf, tmp, size, "Subject: %s...\n\n", aux);
+		free(aux);
+	}
+
+	CONCAT(buf, tmp, size, "%s\n\n", sms->user_data[0].u.text);
+
+	return buf;
+error:
+	if (buf)
+		free(buf);
+	if (aux)
+		free(aux);
+	return NULL;
+}
