@@ -1,6 +1,6 @@
 /*
 
-  $Id: ldif.c,v 1.18 2007/05/24 18:57:47 dforsi Exp $
+  $Id: ldif.c,v 1.24 2007/12/04 19:27:32 hadess Exp $
   
   G N O K I I
 
@@ -30,11 +30,12 @@
 
  */
 
+#include "config.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
-#include "config.h"
 #include "compat.h"
 #include "gnokii.h"
 #include "gnokii-internal.h"
@@ -45,28 +46,27 @@
 */
 static int ldif_entry_write(FILE *f, const char *parameter, const char *value, int convertToUTF8)
 {
-	char *buf;
+	char *buf = NULL;
 	int n, inlen, buflen;
 
 	if (string_base64(value)) {
 		inlen = strlen(value);
-		buf = NULL;
-		n = 3 * inlen;
+		n = 2 * inlen * 4 / 3; /* * 2   -> UTF8
+					* * 4/3 -> base 64
+					*/
+		if (n % 4)
+			n += (4 - (n % 4)); /* base 64 padds with '=' to length dividable by 4 */
 
-		do {
-			buflen = n;
-			if (buf) free(buf);
-			buf = malloc(buflen + 1);
-			if (convertToUTF8)
-				n = utf8_base64_encode(buf, buflen, value, inlen);
-			else
-				n = base64_encode(buf, buflen, value, inlen);
-		} while (n >= buflen);
+		buflen = n;
+		buf = malloc(buflen + 1);
+		if (convertToUTF8)
+			n = utf8_base64_encode(buf, buflen, value, inlen);
+		else
+			n = base64_encode(buf, buflen, value, inlen);
 
 		fprintf(f, "%s:: %s\n", parameter, buf);
 
 		free(buf);
-
 	} else {
 		fprintf(f, "%s: %s\n", parameter, value);
 	}
@@ -97,15 +97,18 @@ GNOKII_API int gn_phonebook2ldif(FILE *f, gn_phonebook_entry *entry)
 	if (entry->caller_group) {
 		char aux2[10];
 
-		memset(aux2, 0, 10);
-		sprintf(aux2, "%d", entry->caller_group);
-		ldif_entry_write(f, "businessGroup", aux2, 1);
+		snprintf(aux2, sizeof(aux2), "%d", entry->caller_group);
+		ldif_entry_write(f, "businessCategory", aux2, 1);
 	}
 
 	if (entry->subentries_count == 0) {
 		ldif_entry_write(f, "telephoneNumber", entry->number, 1);
 	}
 
+	if (entry->address.has_address) {
+		ldif_entry_write(f, "homePostalAddress", entry->address.post_office_box, 1);
+	}
+
 	/* Add ext. pbk info if required */
 	for (i = 0; i < entry->subentries_count; i++) {
 		switch (entry->subentries[i].entry_type) {
@@ -162,14 +165,18 @@ GNOKII_API int gn_phonebook2ldif(FILE *f, gn_phonebook_entry *entry)
 #define STORE(a, b) STORE2(a, b, (void) 0)
 #define STORE_BASE64(a, b) STORE2_BASE64(a, b, (void) 0)
 
-#define STORESUB(a, c) STORE2(a, entry->subentries[entry->subentries_count++].data.number, \
+#define STORESUB(a, c) if (entry->subentries_count == GN_PHONEBOOK_SUBENTRIES_MAX_NUMBER) return -1; \
+				STORE2(a, entry->subentries[entry->subentries_count++].data.number, \
 				entry->subentries[entry->subentries_count].entry_type = c);
-#define STORESUB_BASE64(a, c) STORE2_BASE64(a, entry->subentries[entry->subentries_count++].data.number, \
+#define STORESUB_BASE64(a, c) if (entry->subentries_count == GN_PHONEBOOK_SUBENTRIES_MAX_NUMBER) return -1; \
+				STORE2_BASE64(a, entry->subentries[entry->subentries_count++].data.number, \
 				entry->subentries[entry->subentries_count].entry_type = c);
-#define STORENUM(a, c) STORE2(a, entry->subentries[entry->subentries_count++].data.number, \
+#define STORENUM(a, c) if (entry->subentries_count == GN_PHONEBOOK_SUBENTRIES_MAX_NUMBER) return -1; \
+				STORE2(a, entry->subentries[entry->subentries_count++].data.number, \
 					entry->subentries[entry->subentries_count].entry_type = GN_PHONEBOOK_ENTRY_Number; \
 					entry->subentries[entry->subentries_count].number_type = c);
-#define STORENUM_BASE64(a, c) STORE2_BASE64(a, entry->subentries[entry->subentries_count++].data.number, \
+#define STORENUM_BASE64(a, c) if (entry->subentries_count == GN_PHONEBOOK_SUBENTRIES_MAX_NUMBER) return -1; \
+				STORE2_BASE64(a, entry->subentries[entry->subentries_count++].data.number, \
 					entry->subentries[entry->subentries_count].entry_type = GN_PHONEBOOK_ENTRY_Number; \
 					entry->subentries[entry->subentries_count].number_type = c);
 #undef ERROR
@@ -218,7 +225,7 @@ GNOKII_API int gn_ldif2phonebook(FILE *f, gn_phonebook_entry *entry)
 		STORENUM("telephoneNumber: ", GN_PHONEBOOK_NUMBER_General);
 		STORENUM_BASE64("telephoneNumber:: ", GN_PHONEBOOK_NUMBER_General);
 		
-		STOREINT("callerGroup: ", entry->caller_group);
+		STOREINT("businessCategory: ", entry->caller_group);
 
 		if (BEGINS("\n"))
 			break;
@@ -234,7 +241,7 @@ GNOKII_API int gn_ldif2phonebook(FILE *f, gn_phonebook_entry *entry)
 			case GN_PHONEBOOK_NUMBER_Home:
 			case GN_PHONEBOOK_NUMBER_Mobile:
 				dprintf("setting default number to %s\n", entry->subentries[i].data.number);
-				strncpy(entry->number, entry->subentries[i].data.number, sizeof(entry->number));
+				snprintf(entry->number, sizeof(entry->number), "%s", entry->subentries[i].data.number);
 				break;
 			default:
 				dprintf("unknown type\n");
