@@ -1,6 +1,6 @@
 /*
 
-  $Id: gnapplet.c,v 1.32 2006/04/29 15:06:36 pkot Exp $
+  $Id: gnapplet.c,v 1.35 2007/11/03 22:07:18 bozo Exp $
 
   G N O K I I
 
@@ -35,10 +35,7 @@
 #include <ctype.h>
 #include <assert.h>
 
-#ifdef HAVE_UNISTD_H
-#  include <unistd.h>
-#endif
-
+#include "compat.h"
 #include "misc.h"
 #include "phones/generic.h"
 #include "phones/gnapplet.h"
@@ -118,6 +115,7 @@ static gn_error gnapplet_clock_alarm_read(gn_data *data, struct gn_statemachine
 static gn_error gnapplet_clock_alarm_write(gn_data *data, struct gn_statemachine *state);
 static gn_error gnapplet_profile_read(gn_data *data, struct gn_statemachine *state);
 static gn_error gnapplet_profile_active_read(gn_data *data, struct gn_statemachine *state);
+static gn_error gnapplet_profile_active_write(gn_data *data, struct gn_statemachine *state);
 
 static gn_error gnapplet_incoming_info(int messagetype, unsigned char *message, int length, gn_data *data, struct gn_statemachine *state);
 static gn_error gnapplet_incoming_phonebook(int messagetype, unsigned char *message, int length, gn_data *data, struct gn_statemachine *state);
@@ -502,6 +500,8 @@ static gn_error gnapplet_functions(gn_operation op, gn_data *data, struct gn_sta
 		return gnapplet_profile_read(data, state);
 	case GN_OP_GetActiveProfile:
 		return gnapplet_profile_active_read(data, state);
+	case GN_OP_SetActiveProfile:
+		return gnapplet_profile_active_write(data, state);
 	default:
 		dprintf("gnapplet unimplemented operation: %d\n", op);
 		return GN_ERR_NOTIMPLEMENTED;
@@ -873,7 +873,7 @@ static gn_error gnapplet_incoming_power(int messagetype, unsigned char *message,
 		percent = pkt_get_uint8(&pkt);
 		source = pkt_get_uint8(&pkt);
 		if (error != GN_ERR_NONE) return error;
-		if (data->battery_unit) *data->battery_unit = GN_RF_Percentage;
+		if (data->battery_unit) *data->battery_unit = GN_BU_Percentage;
 		if (data->battery_level) *data->battery_level = percent;
 		if (data->power_source) *data->power_source = source;
 		break;
@@ -1574,6 +1574,19 @@ static gn_error gnapplet_profile_active_read(gn_data *data, struct gn_statemachi
 }
 
 
+static gn_error gnapplet_profile_active_write(gn_data *data, struct gn_statemachine *state)
+{
+	REQUEST_DEF;
+
+	if (!data->profile) return GN_ERR_INTERNALERROR;
+
+	pkt_put_uint16(&pkt, GNAPPLET_MSG_PROFILE_SET_ACTIVE_REQ);
+	pkt_put_uint16(&pkt, data->profile->number);
+
+	SEND_MESSAGE_BLOCK(GNAPPLET_MSG_PROFILE);
+}
+
+
 static gn_error gnapplet_incoming_profile(int messagetype, unsigned char *message, int length, gn_data *data, struct gn_statemachine *state)
 {
 	gn_profile *profile;
@@ -1605,6 +1618,11 @@ static gn_error gnapplet_incoming_profile(int messagetype, unsigned char *messag
 		profile->number = pkt_get_uint16(&pkt);
 		break;
 
+	case GNAPPLET_MSG_PROFILE_SET_ACTIVE_RESP:
+		if (!(profile = data->profile)) return GN_ERR_INTERNALERROR;
+		if (error != GN_ERR_NONE) return error;
+		break;
+
 	default:
 		return GN_ERR_UNHANDLEDFRAME;
 	}
